@@ -6,72 +6,91 @@ from collections import defaultdict
 from itertools import (chain,
                        takewhile)
 
-from reprit.base import generate_repr
-from typing_extensions import final
+from typing_extensions import (Literal,
+                               TypeGuard,
+                               final)
 
 from . import annotated
 
 _MIN_SUB_SIGNATURES_COUNT = 2
 
 
-class Parameter:
-    class Kind(enum.IntEnum):
-        POSITIONAL_ONLY = 0
-        POSITIONAL_OR_KEYWORD = 1
-        VARIADIC_POSITIONAL = 2
-        KEYWORD_ONLY = 3
-        VARIADIC_KEYWORD = 4
+@final
+class _Default(enum.Enum):
+    NONE = enum.auto()
 
-        def __repr__(self) -> str:
-            return type(self).__qualname__ + '.' + self.name
 
-        def __str__(self) -> str:
-            return self.name.lower().replace('_', ' ')
+_NONE = _Default.NONE
 
+
+@final
+class ParameterKind(enum.IntEnum):
+    POSITIONAL_ONLY = 0
+    POSITIONAL_OR_KEYWORD = 1
+    VARIADIC_POSITIONAL = 2
+    KEYWORD_ONLY = 3
+    VARIADIC_KEYWORD = 4
+
+    def __repr__(self) -> str:
+        return type(self).__qualname__ + '.' + self.name
+
+    def __str__(self) -> str:
+        return self.name.lower().replace('_', ' ')
+
+
+class BaseParameter(ABC):
+    @property
+    @abstractmethod
+    def annotation(self) -> t.Any:
+        """Returns annotation of the parameter."""
+
+    @property
+    @abstractmethod
+    def kind(self) -> ParameterKind:
+        """Returns kind of the parameter."""
+
+    @property
+    @abstractmethod
+    def name(self) -> str:
+        """Returns name of the parameter."""
+
+
+@final
+class RequiredParameter(BaseParameter):
     @property
     def annotation(self) -> t.Any:
         return self._annotation
 
     @property
-    def has_default(self) -> bool:
-        return self._has_default
-
-    @property
-    def kind(self) -> Kind:
+    def kind(self) -> ParameterKind:
         return self._kind
 
     @property
     def name(self) -> str:
         return self._name
 
-    __slots__ = '_annotation', '_has_default', '_kind', '_name'
+    __slots__ = '_annotation', '_kind', '_name'
 
     _annotation: t.Any
-    _has_default: bool
-    _kind: Kind
+    _kind: ParameterKind
     _name: str
 
     def __new__(cls,
                 *,
                 annotation: t.Any,
-                has_default: bool,
-                kind: Kind,
-                name: str) -> 'Parameter':
-        # performing validation inside of `__init__` instead of `__new__`,
-        # because `pickle` does not support keyword only arguments in `__new__`
-        if ((kind is cls.Kind.VARIADIC_POSITIONAL
-             or kind is cls.Kind.VARIADIC_KEYWORD)
-                and has_default):
-            raise ValueError('Variadic parameters '
-                             'can\'t have default arguments.')
+                kind: Literal[ParameterKind.POSITIONAL_ONLY,
+                              ParameterKind.POSITIONAL_OR_KEYWORD,
+                              ParameterKind.KEYWORD_ONLY],
+                name: str) -> 'RequiredParameter':
+        if (kind is ParameterKind.VARIADIC_POSITIONAL
+                or kind is ParameterKind.VARIADIC_KEYWORD):
+            raise ValueError('Variadic parameters can\'t be required.')
         self = super().__new__(cls)
-        self._annotation, self._has_default, self._kind, self._name = (
-            annotation, has_default, kind, name
-        )
+        self._annotation, self._kind, self._name = annotation, kind, name
         return self
 
     @t.overload
-    def __eq__(self, other: 'Parameter') -> bool:
+    def __eq__(self, other: BaseParameter) -> bool:
         ...
 
     @t.overload
@@ -79,40 +98,143 @@ class Parameter:
         ...
 
     def __eq__(self, other):
-        return ((self.name == other.name
+        return ((type(self) is type(other)
+                 and self.name == other.name
                  and self.kind is other.kind
-                 and self.has_default is other.has_default
                  and annotated.are_equal(self.annotation, other.annotation))
-                if isinstance(other, Parameter)
+                if isinstance(other, BaseParameter)
                 else NotImplemented)
 
     def __getnewargs_ex__(self) -> t.Tuple[t.Tuple[()], t.Dict[str, t.Any]]:
         return (), {'annotation': self._annotation,
-                    'has_default': self._has_default,
                     'kind': self._kind,
                     'name': self._name}
 
     def __hash__(self) -> int:
-        return hash((self._annotation, self._has_default, self._kind,
-                     self._name))
+        return hash((self._annotation, self._kind, self._name))
 
-    __repr__ = generate_repr(__new__,
-                             argument_serializer=annotated.to_repr)
+    def __repr__(self) -> str:
+        return (f'{type(self).__qualname__}('
+                f'annotation={annotated.to_repr(self.annotation)}, '
+                f'kind={repr(self.kind)}, '
+                f'name={repr(self.name)}'
+                f')')
 
     def __str__(self) -> str:
-        return ''.join(['*'
-                        if self.kind is self.Kind.VARIADIC_POSITIONAL
-                        else ('**'
-                              if self.kind is self.Kind.VARIADIC_KEYWORD
-                              else ''),
+        return f'{self.name}: {annotated.to_repr(self.annotation)}'
+
+
+@final
+class OptionalParameter(BaseParameter):
+    @property
+    def annotation(self) -> t.Any:
+        return self._annotation
+
+    @property
+    def default(self) -> t.Any:
+        return self._default
+
+    @property
+    def has_default(self) -> bool:
+        return self._has_default
+
+    @property
+    def kind(self) -> ParameterKind:
+        return self._kind
+
+    @property
+    def name(self) -> str:
+        return self._name
+
+    __slots__ = '_annotation', '_default', '_has_default', '_kind', '_name'
+
+    _annotation: t.Any
+    _default: t.Union[_Default, t.Any]
+    _has_default: bool
+    _kind: ParameterKind
+    _name: str
+
+    def __new__(cls,
+                *,
+                annotation: t.Any,
+                default: t.Union[_Default, t.Any] = _NONE,
+                kind: ParameterKind,
+                name: str) -> 'OptionalParameter':
+        has_default = default is not _NONE
+        if ((kind is ParameterKind.VARIADIC_POSITIONAL
+             or kind is ParameterKind.VARIADIC_KEYWORD)
+                and has_default):
+            raise ValueError('Variadic parameters '
+                             'can\'t have default arguments.')
+        self = super().__new__(cls)
+        (
+            self._annotation, self._has_default, self._kind, self._name
+        ) = annotation, has_default, kind, name
+        if has_default:
+            self._default = default
+        return self
+
+    @t.overload
+    def __eq__(self, other: BaseParameter) -> bool:
+        ...
+
+    @t.overload
+    def __eq__(self, other: t.Any) -> t.Any:
+        ...
+
+    def __eq__(self, other):
+        return ((type(self) is type(other)
+                 and self.name == other.name
+                 and self.kind is other.kind
+                 and self.has_default is other.has_default
+                 and (not self.has_default
+                      or annotated.are_equal(self.default, other.default))
+                 and annotated.are_equal(self.annotation, other.annotation))
+                if isinstance(other, BaseParameter)
+                else NotImplemented)
+
+    def __getnewargs_ex__(self) -> t.Tuple[t.Tuple[()], t.Dict[str, t.Any]]:
+        return (), {'annotation': self._annotation,
+                    **({'default': self._default}
+                       if self._has_default
+                       else {}),
+                    'kind': self._kind,
+                    'name': self._name}
+
+    def __hash__(self) -> int:
+        return hash((self._annotation,
+                     *((self._default,) if self._has_default else ()),
+                     self._kind, self._name))
+
+    def __repr__(self) -> str:
+        return (f'{type(self).__qualname__}('
+                f'annotation={annotated.to_repr(self.annotation)}, '
+                + (f'default={annotated.to_repr(self.default)}, '
+                   if self.has_default
+                   else '')
+                + f'kind={repr(self.kind)}, '
+                  f'name={repr(self.name)}'
+                  f')')
+
+    def __str__(self) -> str:
+        return ''.join([('*'
+                         if self.kind is ParameterKind.VARIADIC_POSITIONAL
+                         else ('**'
+                               if self.kind is ParameterKind.VARIADIC_KEYWORD
+                               else '')),
                         self.name,
                         f': {annotated.to_repr(self.annotation)}',
-                        ' = ...' if self.has_default else ''])
+                        (f' = {annotated.to_repr(self.default)}'
+                         if self.has_default
+                         else '')])
+
+
+Parameter = t.Union[OptionalParameter, RequiredParameter]
 
 
 def to_parameters_by_kind(
         parameters: t.Iterable[Parameter]
-) -> t.Dict[Parameter.Kind, t.List[Parameter]]:
+) -> t.Dict[ParameterKind, t.List[Parameter]]:
     result = defaultdict(list)
     for parameter in parameters:
         result[parameter.kind].append(parameter)
@@ -125,10 +247,11 @@ def to_parameters_by_name(
     return {parameter.name: parameter for parameter in parameters}
 
 
-def all_parameters_has_defaults(
+def all_parameters_are_optional(
         parameters: t.Iterable[Parameter]
-) -> bool:
-    return all(parameter.has_default for parameter in parameters)
+) -> TypeGuard[t.Iterable[OptionalParameter]]:
+    return all(isinstance(parameter, OptionalParameter)
+               for parameter in parameters)
 
 
 _Self = t.TypeVar('_Self')
@@ -171,11 +294,11 @@ class PlainSignature(BaseSignature):
     def all_set(self, *args: _Arg, **kwargs: _KwArg) -> bool:
         parameters_by_kind = to_parameters_by_kind(self._parameters)
         positionals = (
-                parameters_by_kind[Parameter.Kind.POSITIONAL_ONLY]
-                + parameters_by_kind[Parameter.Kind.POSITIONAL_OR_KEYWORD]
+                parameters_by_kind[ParameterKind.POSITIONAL_ONLY]
+                + parameters_by_kind[ParameterKind.POSITIONAL_OR_KEYWORD]
         )
         unexpected_positional_arguments_found = (
-                not parameters_by_kind[Parameter.Kind.VARIADIC_POSITIONAL]
+                not parameters_by_kind[ParameterKind.VARIADIC_POSITIONAL]
                 and len(args) > len(positionals)
         )
         if unexpected_positional_arguments_found:
@@ -183,12 +306,12 @@ class PlainSignature(BaseSignature):
         rest_positionals = positionals[len(args):]
         rest_positionals_by_kind = to_parameters_by_kind(rest_positionals)
         rest_keywords: t.Iterable[Parameter] = (
-                rest_positionals_by_kind[Parameter.Kind.POSITIONAL_OR_KEYWORD]
-                + parameters_by_kind[Parameter.Kind.KEYWORD_ONLY]
+                rest_positionals_by_kind[ParameterKind.POSITIONAL_OR_KEYWORD]
+                + parameters_by_kind[ParameterKind.KEYWORD_ONLY]
         )
         rest_keywords_by_name = to_parameters_by_name(rest_keywords)
         unexpected_keyword_arguments_found = (
-                not parameters_by_kind[Parameter.Kind.VARIADIC_KEYWORD]
+                not parameters_by_kind[ParameterKind.VARIADIC_KEYWORD]
                 and kwargs.keys() - rest_keywords_by_name.keys()
         )
         if unexpected_keyword_arguments_found:
@@ -199,20 +322,20 @@ class PlainSignature(BaseSignature):
             if name not in kwargs
         }
         rest_positionals_only = rest_positionals_by_kind[
-            Parameter.Kind.POSITIONAL_ONLY
+            ParameterKind.POSITIONAL_ONLY
         ]
         rest_keywords = rest_keywords_by_name.values()
-        return (all_parameters_has_defaults(rest_positionals_only)
-                and all_parameters_has_defaults(rest_keywords))
+        return (all_parameters_are_optional(rest_positionals_only)
+                and all_parameters_are_optional(rest_keywords))
 
     def expects(self, *args: _Arg, **kwargs: _KwArg) -> bool:
         parameters_by_kind = to_parameters_by_kind(self._parameters)
-        positionals = (parameters_by_kind[Parameter.Kind.POSITIONAL_ONLY]
+        positionals = (parameters_by_kind[ParameterKind.POSITIONAL_ONLY]
                        + parameters_by_kind[
-                           Parameter.Kind.POSITIONAL_OR_KEYWORD
+                           ParameterKind.POSITIONAL_OR_KEYWORD
                        ])
         unexpected_positional_arguments_found = (
-                not parameters_by_kind[Parameter.Kind.VARIADIC_POSITIONAL]
+                not parameters_by_kind[ParameterKind.VARIADIC_POSITIONAL]
                 and len(args) > len(positionals)
         )
         if unexpected_positional_arguments_found:
@@ -220,12 +343,12 @@ class PlainSignature(BaseSignature):
         rest_positionals = positionals[len(args):]
         rest_positionals_by_kind = to_parameters_by_kind(rest_positionals)
         rest_keywords = (
-                rest_positionals_by_kind[Parameter.Kind.POSITIONAL_OR_KEYWORD]
-                + parameters_by_kind[Parameter.Kind.KEYWORD_ONLY]
+                rest_positionals_by_kind[ParameterKind.POSITIONAL_OR_KEYWORD]
+                + parameters_by_kind[ParameterKind.KEYWORD_ONLY]
         )
         rest_keywords_by_name = to_parameters_by_name(rest_keywords)
         unexpected_keyword_arguments_found = (
-                not parameters_by_kind[Parameter.Kind.VARIADIC_KEYWORD]
+                not parameters_by_kind[ParameterKind.VARIADIC_KEYWORD]
                 and kwargs.keys() - rest_keywords_by_name.keys()
         )
         return not unexpected_keyword_arguments_found
@@ -238,13 +361,13 @@ class PlainSignature(BaseSignature):
                                 self._parameters, args, kwargs,
                                 has_variadic=bool(
                                         parameters_by_kind[
-                                            Parameter.Kind.VARIADIC_POSITIONAL
+                                            ParameterKind.VARIADIC_POSITIONAL
                                         ]
                                 )
                         ),
                         kwargs,
                         has_variadic=bool(parameters_by_kind[
-                                              Parameter.Kind.VARIADIC_KEYWORD
+                                              ParameterKind.VARIADIC_KEYWORD
                                           ])
                 ),
                 returns=self._returns
@@ -278,18 +401,18 @@ class PlainSignature(BaseSignature):
                                      f'with kind "{prior.kind!s}" '
                                      f'precedes parameter "{name}" '
                                      f'with kind "{kind!s}".')
-                if (kind is Parameter.Kind.POSITIONAL_OR_KEYWORD
-                        or kind is Parameter.Kind.POSITIONAL_ONLY):
-                    if not parameter.has_default:
-                        if prior.has_default:
-                            raise ValueError('Invalid parameters order: '
-                                             f'parameter "{name}" '
-                                             'without default argument '
-                                             'follows '
-                                             f'parameter "{prior.name}" '
-                                             'with default argument.')
-                elif (kind is not Parameter.Kind.POSITIONAL_OR_KEYWORD
-                      and kind is not Parameter.Kind.KEYWORD_ONLY):
+                if (kind is ParameterKind.POSITIONAL_OR_KEYWORD
+                        or kind is ParameterKind.POSITIONAL_ONLY):
+                    if (isinstance(parameter, RequiredParameter)
+                            and isinstance(prior, OptionalParameter)):
+                        raise ValueError(
+                                'Invalid parameters order: '
+                                f'required positional parameter "{name}" '
+                                'follows '
+                                f'optional parameter "{prior.name}".'
+                        )
+                elif (kind is ParameterKind.VARIADIC_POSITIONAL
+                      or kind is ParameterKind.VARIADIC_KEYWORD):
                     if kind in visited_kinds:
                         raise ValueError('Variadic parameters '
                                          'should have unique kinds, '
@@ -324,29 +447,33 @@ class PlainSignature(BaseSignature):
     def __hash__(self) -> int:
         return hash((self._parameters, self.returns))
 
-    __repr__ = generate_repr(__new__,
-                             argument_serializer=annotated.to_repr)
+    def __repr__(self) -> str:
+        return (f'{type(self).__qualname__}('
+                + (f'{", ".join(map(repr, self._parameters))}, '
+                   if self._parameters
+                   else '')
+                + f'returns={annotated.to_repr(self._returns)})')
 
     def __str__(self) -> str:
         parameters_by_kind = to_parameters_by_kind(self._parameters)
-        positionals_only = parameters_by_kind[Parameter.Kind.POSITIONAL_ONLY]
+        positionals_only = parameters_by_kind[ParameterKind.POSITIONAL_ONLY]
         parts = list(map(str, positionals_only))
         if positionals_only:
             parts.append(self.POSITIONAL_ONLY_SEPARATOR)
         parts.extend(map(str,
                          parameters_by_kind[
-                             Parameter.Kind.POSITIONAL_OR_KEYWORD
+                             ParameterKind.POSITIONAL_OR_KEYWORD
                          ]))
         variadic_positionals = parameters_by_kind[
-            Parameter.Kind.VARIADIC_POSITIONAL
+            ParameterKind.VARIADIC_POSITIONAL
         ]
         parts.extend(map(str, variadic_positionals))
-        keywords_only = parameters_by_kind[Parameter.Kind.KEYWORD_ONLY]
+        keywords_only = parameters_by_kind[ParameterKind.KEYWORD_ONLY]
         if keywords_only and not variadic_positionals:
             parts.append(self.KEYWORD_ONLY_SEPARATOR)
         parts.extend(map(str, keywords_only))
         parts.extend(map(str,
-                         parameters_by_kind[Parameter.Kind.VARIADIC_KEYWORD]))
+                         parameters_by_kind[ParameterKind.VARIADIC_KEYWORD]))
         return ('(' + ', '.join(parts) + ') -> '
                 + annotated.to_repr(self.returns))
 
@@ -408,7 +535,9 @@ class OverloadedSignature(BaseSignature):
     def __hash__(self) -> int:
         return hash(self._signatures)
 
-    __repr__ = generate_repr(__new__)
+    def __repr__(self) -> str:
+        return (f'{type(self).__qualname__}'
+                f'({", ".join(map(repr, self._signatures))})')
 
     def __str__(self) -> str:
         return ' | '.join(map(str, self._signatures))
@@ -426,8 +555,8 @@ def _bind_positionals(parameters: t.Tuple[Parameter, ...],
                       *,
                       has_variadic: bool) -> t.Tuple[Parameter, ...]:
     def is_positionable(parameter: Parameter) -> bool:
-        return (parameter.kind is Parameter.Kind.POSITIONAL_OR_KEYWORD
-                or parameter.kind is Parameter.Kind.POSITIONAL_ONLY)
+        return (parameter.kind is ParameterKind.POSITIONAL_OR_KEYWORD
+                or parameter.kind is ParameterKind.POSITIONAL_ONLY)
 
     positionals = tuple(takewhile(is_positionable, parameters))
     if len(args) > len(positionals) and not has_variadic:
@@ -438,8 +567,8 @@ def _bind_positionals(parameters: t.Tuple[Parameter, ...],
     for positional in positionals[:len(args)]:
         if positional.name in kwargs:
             kind = positional.kind
-            if (kind is Parameter.Kind.POSITIONAL_OR_KEYWORD
-                    or kind is Parameter.Kind.KEYWORD_ONLY):
+            if (kind is ParameterKind.POSITIONAL_OR_KEYWORD
+                    or kind is ParameterKind.KEYWORD_ONLY):
                 raise TypeError('Got multiple values '
                                 f'for parameter "{positional.name}".')
             else:
@@ -457,8 +586,8 @@ def _bind_keywords(parameters: t.Tuple[Parameter, ...],
             kwargs_names
             - {parameter.name
                for parameter in parameters
-               if (parameter.kind is Parameter.Kind.POSITIONAL_OR_KEYWORD
-                   or parameter.kind is Parameter.Kind.KEYWORD_ONLY)}
+               if (parameter.kind is ParameterKind.POSITIONAL_OR_KEYWORD
+                   or parameter.kind is ParameterKind.KEYWORD_ONLY)}
     )
     if extra_kwargs_names and not has_variadic:
         value = 'argument' + 's' * (len(extra_kwargs_names) != 1)
@@ -472,17 +601,27 @@ def _bind_keywords(parameters: t.Tuple[Parameter, ...],
                              if parameter.name in kwargs_names)
     return (parameters[:first_kwarg_index]
             + tuple(
-                    Parameter(
-                            annotation=parameter.annotation,
-                            name=parameter.name,
-                            kind=Parameter.Kind.KEYWORD_ONLY,
-                            has_default=(parameter.has_default
-                                         or parameter.name in kwargs_names)
+                    (
+                        OptionalParameter(
+                                annotation=parameter.annotation,
+                                name=parameter.name,
+                                kind=ParameterKind.KEYWORD_ONLY,
+                                **({'default': parameter.default}
+                                   if (isinstance(parameter, OptionalParameter)
+                                       and parameter.has_default)
+                                   else ({'default': kwargs[parameter.name]}
+                                         if parameter.name in kwargs_names
+                                         else {}))
+                        )
+                        if (isinstance(parameter, OptionalParameter)
+                            or parameter.name in kwargs_names)
+                        else RequiredParameter(annotation=parameter.annotation,
+                                               name=parameter.name,
+                                               kind=ParameterKind.KEYWORD_ONLY)
                     )
-                    if (parameter.kind is Parameter.Kind.POSITIONAL_OR_KEYWORD
-                        or parameter.kind is Parameter.Kind.KEYWORD_ONLY)
+                    if (parameter.kind is ParameterKind.POSITIONAL_OR_KEYWORD
+                        or parameter.kind is ParameterKind.KEYWORD_ONLY)
                     else parameter
                     for parameter in parameters[first_kwarg_index:]
-                    if (parameter.kind
-                        is not Parameter.Kind.VARIADIC_POSITIONAL)
+                    if parameter.kind is not ParameterKind.VARIADIC_POSITIONAL
             ))
