@@ -1,8 +1,10 @@
+import types
 import typing as t
 from functools import reduce
 
 from . import catalog
 
+ModuleOrType = t.Union[types.ModuleType, type]
 Namespace = t.Dict[str, t.Any]
 
 
@@ -10,14 +12,8 @@ class ObjectNotFound(Exception):
     pass
 
 
-def search(namespace: Namespace, object_path: catalog.Path) -> t.Any:
+def search(module_or_type: ModuleOrType, object_path: catalog.Path) -> t.Any:
     try:
-        root_object = namespace[object_path[0]]
-    except KeyError:
-        raise ObjectNotFound(object_path)
-    try:
-        return (reduce(getattr, object_path[1:], root_object)
-                if len(object_path) > 1
-                else root_object)
+        return reduce(getattr, object_path, module_or_type)
     except AttributeError:
         raise ObjectNotFound(object_path)
