@@ -2,9 +2,11 @@ import typing as _t
 from ._core import (models as _models,
                     signatures as _signatures)
 
+OptionalParameter = _models.OptionalParameter
 OverloadedSignature = _models.OverloadedSignature
-Parameter = _models.Parameter
+ParameterKind = _models.ParameterKind
 PlainSignature = _models.PlainSignature
+RequiredParameter = _models.RequiredParameter
 
 
 def signature_from_callable(
